@@ -1,4 +1,4 @@
-""" `SNPs` reads, writes, and remaps genotype / raw data files.
+""" ``SNPs`` reads, writes, merges, and remaps genotype / raw data files.
 
 """
 
@@ -39,6 +39,7 @@ from itertools import groupby, count
 import logging
 import os
 import re
+import warnings
 
 import numpy as np
 import pandas as pd
@@ -46,8 +47,8 @@ from pandas.api.types import CategoricalDtype
 
 from snps.ensembl import EnsemblRestClient
 from snps.resources import Resources
-from snps.io import Reader, Writer
-from snps.utils import Parallelizer, get_empty_snps_dataframe
+from snps.io import Reader, Writer, get_empty_snps_dataframe
+from snps.utils import Parallelizer
 
 logger = logging.getLogger(__name__)
 
@@ -57,11 +58,12 @@ class SNPs:
         self,
         file="",
         only_detect_source=False,
-        assign_par_snps=True,
+        assign_par_snps=False,
         output_dir="output",
         resources_dir="resources",
         deduplicate=True,
         deduplicate_XY_chrom=True,
+        deduplicate_MT_chrom=True,
         parallelize=False,
         processes=os.cpu_count(),
         rsids=(),
@@ -81,9 +83,12 @@ class SNPs:
         resources_dir : str
             name / path of resources directory
         deduplicate : bool
-            deduplicate RSIDs and make SNPs available as `duplicate_snps`
+            deduplicate RSIDs and make SNPs available as `SNPs.duplicate`
         deduplicate_XY_chrom : bool
-            deduplicate alleles in the non-PAR regions of X and Y for males; see `discrepant_XY_snps`
+            deduplicate alleles in the non-PAR regions of X and Y for males; see
+            `SNPs.discrepant_XY`
+        deduplicate_MT_chrom : bool
+            deduplicate alleles on MT; see `SNPs.heterozygous_MT`
         parallelize : bool
             utilize multiprocessing to speedup calculations
         processes : int
@@ -94,9 +99,13 @@ class SNPs:
         self._file = file
         self._only_detect_source = only_detect_source
         self._snps = get_empty_snps_dataframe()
-        self._duplicate_snps = pd.DataFrame()
-        self._discrepant_XY_snps = pd.DataFrame()
-        self._source = ""
+        self._duplicate = get_empty_snps_dataframe()
+        self._discrepant_XY = get_empty_snps_dataframe()
+        self._heterozygous_MT = get_empty_snps_dataframe()
+        self._discrepant_vcf_position = get_empty_snps_dataframe()
+        self._discrepant_merge_positions = pd.DataFrame()
+        self._discrepant_merge_genotypes = pd.DataFrame()
+        self._source = []
         self._phased = False
         self._build = 0
         self._build_detected = False
@@ -118,19 +127,24 @@ class SNPs:
             d["snps"].rename(index=multi_rsids, inplace=True)
 
             self._snps = d["snps"]
-            self._source = d["source"]
+            self._source = (
+                d["source"].split(", ") if ", " in d["source"] else [d["source"]]
+            )
             self._phased = d["phased"]
-            self._build = d["build"]
-            self._build_detected = True if d["build"] else False
 
             if not self._snps.empty:
-                self.sort_snps()
+                self.sort()
 
                 if deduplicate:
                     self._deduplicate_rsids()
 
+                # prefer to use SNP positions to detect build
+                self._build = self.detect_build()
+                self._build_detected = True if self._build else False
+
                 if not self._build_detected:
-                    self._build = self.detect_build()
+                    # use build detected from `read` method or comments, if any
+                    self._build = d["build"]
 
                     if not self._build:
                         self._build = 37  # assume Build 37 / GRCh37 if not detected
@@ -139,11 +153,14 @@ class SNPs:
 
                 if assign_par_snps:
                     self._assign_par_snps()
-                    self.sort_snps()
+                    self.sort()
 
                 if deduplicate_XY_chrom:
                     if self.determine_sex() == "Male":
                         self._deduplicate_XY_chrom()
+
+                if deduplicate_MT_chrom:
+                    self._deduplicate_MT_chrom()
             else:
                 logger.warning("no SNPs loaded...")
 
@@ -152,27 +169,49 @@ class SNPs:
 
     @property
     def source(self):
-        """ Summary of the SNP data source for ``SNPs``.
+        """ Summary of the SNP data source(s).
 
         Returns
         -------
         str
+            Data source(s) for this ``SNPs`` object, separated by ", ".
         """
-        return self._source
+        return ", ".join(self._source)
 
     @property
     def snps(self):
-        """ Get a copy of SNPs.
+        """ Normalized SNPs.
+
+        Notes
+        -----
+        Throughout ``snps``, the "normalized ``snps`` dataframe" is defined as follows:
+
+        =============  ===================================  ===============
+        Column         Description                          `pandas` dtype
+        =============  ===================================  ===============
+        rsid [*]_      SNP ID                               object (string)
+        chrom          Chromosome of SNP                    object (string)
+        pos            Position of SNP (relative to build)  uint32
+        genotype [*]_  Genotype of SNP                      object (string)
+        =============  ===================================  ===============
+
+        .. [*] Dataframe index
+        .. [*] Genotype can be null, length 1, or length 2. Specifically, genotype is null if not
+          called or unavailable. Otherwise, for autosomal chromosomes, genotype is two alleles.
+          For the X and Y chromosomes, male genotypes are one allele in the non-PAR regions
+          (assuming `deduplicate_XY_chrom`). For the MT chromosome, genotypes are one allele
+          (assuming `deduplicate_MT_chrom`).
 
         Returns
         -------
         pandas.DataFrame
+            normalized ``snps`` dataframe
         """
         return self._snps
 
     @property
-    def duplicate_snps(self):
-        """ Get any duplicate SNPs.
+    def duplicate(self):
+        """ Duplicate SNPs.
 
         A duplicate SNP has the same RSID as another SNP. The first occurrence
         of the RSID is not considered a duplicate SNP.
@@ -180,12 +219,13 @@ class SNPs:
         Returns
         -------
         pandas.DataFrame
+            normalized ``snps`` dataframe
         """
-        return self._duplicate_snps
+        return self._duplicate
 
     @property
-    def discrepant_XY_snps(self):
-        """ Get any discrepant XY SNPs.
+    def discrepant_XY(self):
+        """ Discrepant XY SNPs.
 
         A discrepant XY SNP is a heterozygous SNP in the non-PAR region of the X
         or Y chromosome found during deduplication for a detected male genotype.
@@ -193,12 +233,117 @@ class SNPs:
         Returns
         -------
         pandas.DataFrame
+            normalized ``snps`` dataframe
         """
-        return self._discrepant_XY_snps
+        return self._discrepant_XY
+
+    @property
+    def heterozygous_MT(self):
+        """ Heterozygous SNPs on the MT chromosome found during deduplication.
+
+        Returns
+        -------
+        pandas.DataFrame
+            normalized ``snps`` dataframe
+        """
+        return self._heterozygous_MT
+
+    @property
+    def discrepant_vcf_position(self):
+        """ SNPs with discrepant positions discovered while saving VCF.
+
+        Returns
+        -------
+        pandas.DataFrame
+            normalized ``snps`` dataframe
+        """
+        return self._discrepant_vcf_position
+
+    @property
+    def discrepant_merge_positions(self):
+        """ SNPs with discrepant positions discovered while merging SNPs.
+
+        Notes
+        -----
+        Definitions of columns in this dataframe are as follows:
+
+        ==============  ===========
+        Column          Description
+        ==============  ===========
+        rsid            SNP ID
+        chrom           Chromosome of existing SNP
+        pos             Position of existing SNP
+        genotype        Genotype of existing SNP
+        chrom_added     Chromosome of added SNP
+        pos_added       Position of added SNP (discrepant with pos)
+        genotype_added  Genotype of added SNP
+        ==============  ===========
+
+        Returns
+        -------
+        pandas.DataFrame
+        """
+        return self._discrepant_merge_positions
+
+    @property
+    def discrepant_merge_genotypes(self):
+        """ SNPs with discrepant genotypes discovered while merging SNPs.
+
+        Notes
+        -----
+        Definitions of columns in this dataframe are as follows:
+
+        ===============  ===========
+        Column           Description
+        ===============  ===========
+        rsid             SNP ID
+        chrom            Chromosome of existing SNP
+        pos              Position of existing SNP
+        genotype         Genotype of existing SNP
+        chrom_added      Chromosome of added SNP
+        pos_added        Position of added SNP
+        genotype_added   Genotype of added SNP (discrepant with genotype)
+        ===============  ===========
+
+
+        Returns
+        -------
+        pandas.DataFrame
+        """
+        return self._discrepant_merge_genotypes
+
+    @property
+    def discrepant_merge_positions_genotypes(self):
+        """ SNPs with discrepant positions and / or genotypes discovered while merging SNPs.
+
+        Notes
+        -----
+        Definitions of columns in this dataframe are as follows:
+
+        ===============  ===========
+        Column           Description
+        ===============  ===========
+        rsid             SNP ID
+        chrom            Chromosome of existing SNP
+        pos              Position of existing SNP
+        genotype         Genotype of existing SNP
+        chrom_added      Chromosome of added SNP
+        pos_added        Position of added SNP (possibly discrepant with pos)
+        genotype_added   Genotype of added SNP (possibly discrepant with genotype)
+        ===============  ===========
+
+        Returns
+        -------
+        pandas.DataFrame
+        """
+        df = self._discrepant_merge_positions.append(self._discrepant_merge_genotypes)
+        if len(df) > 1:
+            df = df.drop_duplicates()
+        return df
 
     @property
     def build(self):
-        """ Get the build of ``SNPs``.
+        """ Build of SNPs.
 
         Returns
         -------
@@ -208,7 +353,7 @@ class SNPs:
 
     @property
     def build_detected(self):
-        """ Get status indicating if build of ``SNPs`` was detected.
+        """ Status indicating if build of SNPs was detected.
 
         Returns
         -------
@@ -218,49 +363,85 @@ class SNPs:
 
     @property
     def assembly(self):
-        """ Get the assembly of ``SNPs``.
+        """ Assembly of SNPs.
 
         Returns
         -------
         str
         """
-        return self.get_assembly()
+        if self.build == 37:
+            return "GRCh37"
+        elif self.build == 36:
+            return "NCBI36"
+        elif self.build == 38:
+            return "GRCh38"
+        else:
+            return ""
 
     @property
-    def snp_count(self):
+    def count(self):
         """ Count of SNPs.
 
         Returns
         -------
         int
         """
-        return self.get_snp_count()
+        return self.get_count()
 
     @property
     def chromosomes(self):
-        """ Chromosomes of ``SNPs``.
+        """ Chromosomes of SNPs.
 
         Returns
         -------
         list
             list of str chromosomes (e.g., ['1', '2', '3', 'MT'], empty list if no chromosomes
         """
-        return self.get_chromosomes()
+        if not self._snps.empty:
+            return list(pd.unique(self._snps["chrom"]))
+        else:
+            return []
 
     @property
     def chromosomes_summary(self):
-        """ Summary of the chromosomes of ``SNPs``.
+        """ Summary of the chromosomes of SNPs.
 
         Returns
         -------
         str
             human-readable listing of chromosomes (e.g., '1-3, MT'), empty str if no chromosomes
         """
-        return self.get_chromosomes_summary()
+        if not self._snps.empty:
+            chroms = list(pd.unique(self._snps["chrom"]))
+
+            int_chroms = [int(chrom) for chrom in chroms if chrom.isdigit()]
+            str_chroms = [chrom for chrom in chroms if not chrom.isdigit()]
+
+            # https://codereview.stackexchange.com/a/5202
+            def as_range(iterable):
+                l = list(iterable)
+                if len(l) > 1:
+                    return "{0}-{1}".format(l[0], l[-1])
+                else:
+                    return "{0}".format(l[0])
+
+            # create str representations
+            int_chroms = ", ".join(
+                as_range(g)
+                for _, g in groupby(int_chroms, key=lambda n, c=count(): n - next(c))
+            )
+            str_chroms = ", ".join(str_chroms)
+
+            if int_chroms != "" and str_chroms != "":
+                int_chroms += ", "
+
+            return int_chroms + str_chroms
+        else:
+            return ""
 
     @property
     def sex(self):
-        """ Sex derived from ``SNPs``.
+        """ Sex derived from SNPs.
 
         Returns
         -------
@@ -280,7 +461,7 @@ class SNPs:
         -------
         bool
         """
-        if self.snp_count == 0 and self.source == "vcf":
+        if self.count == 0 and self.source == "vcf":
             return True
 
         return False
@@ -295,7 +476,7 @@ class SNPs:
         """
         return self._phased
 
-    def heterozygous_snps(self, chrom=""):
+    def heterozygous(self, chrom=""):
         """ Get heterozygous SNPs.
 
         Parameters
@@ -306,6 +487,7 @@ class SNPs:
         Returns
         -------
         pandas.DataFrame
+            normalized ``snps`` dataframe
         """
         if chrom:
             return self._snps.loc[
@@ -321,8 +503,8 @@ class SNPs:
                 & (self._snps.genotype.str[0] != self._snps.genotype.str[1])
             ]
 
-    def not_null_snps(self, chrom=""):
-        """ Get not null SNPs.
+    def homozygous(self, chrom=""):
+        """ Get homozygous SNPs.
 
         Parameters
         ----------
@@ -332,6 +514,34 @@ class SNPs:
         Returns
         -------
         pandas.DataFrame
+            normalized ``snps`` dataframe
+        """
+        if chrom:
+            return self._snps.loc[
+                (self._snps.chrom == chrom)
+                & (self._snps.genotype.notnull())
+                & (self._snps.genotype.str.len() == 2)
+                & (self._snps.genotype.str[0] == self._snps.genotype.str[1])
+            ]
+        else:
+            return self._snps.loc[
+                (self._snps.genotype.notnull())
+                & (self._snps.genotype.str.len() == 2)
+                & (self._snps.genotype.str[0] == self._snps.genotype.str[1])
+            ]
+
+    def notnull(self, chrom=""):
+        """ Get not null genotype SNPs.
+
+        Parameters
+        ----------
+        chrom : str, optional
+            chromosome (e.g., "1", "X", "MT")
+
+        Returns
+        -------
+        pandas.DataFrame
+            normalized ``snps`` dataframe
         """
 
         if chrom:
@@ -341,15 +551,16 @@ class SNPs:
         else:
             return self._snps.loc[self._snps.genotype.notnull()]
 
-    def get_summary(self):
-        """ Get summary of ``SNPs``.
+    @property
+    def summary(self):
+        """ Summary of SNPs.
 
         Returns
         -------
         dict
             summary info if ``SNPs`` is valid, else {}
         """
-        if not self.is_valid():
+        if not self.valid:
             return {}
         else:
             return {
@@ -357,12 +568,13 @@ class SNPs:
                 "assembly": self.assembly,
                 "build": self.build,
                 "build_detected": self.build_detected,
-                "snp_count": self.snp_count,
+                "count": self.count,
                 "chromosomes": self.chromosomes_summary,
                 "sex": self.sex,
             }
 
-    def is_valid(self):
+    @property
+    def valid(self):
         """ Determine if ``SNPs`` is valid.
 
         ``SNPs`` is valid when the input file has been successfully parsed.
@@ -372,12 +584,12 @@ class SNPs:
         bool
             True if ``SNPs`` is valid
         """
-        if self._snps.empty:
+        if self.snps.empty:
             return False
         else:
             return True
 
-    def save_snps(self, filename="", vcf=False, atomic=True, **kwargs):
+    def save(self, filename="", vcf=False, atomic=True, **kwargs):
         """ Save SNPs to file.
 
         Parameters
@@ -395,11 +607,30 @@ class SNPs:
         -------
         str
             path to file in output directory if SNPs were saved, else empty str
+
+        References
+        ----------
+        1. Fluent Python by Luciano Ramalho (O'Reilly). Copyright 2015 Luciano Ramalho,
+           978-1-491-94600-8.
         """
-        return Writer.write_file(
+        if "sep" not in kwargs:
+            kwargs["sep"] = "\t"
+
+        path, *extra = Writer.write_file(
             snps=self, filename=filename, vcf=vcf, atomic=atomic, **kwargs
         )
 
+        if len(extra) == 1 and not extra[0].empty:
+            self._discrepant_vcf_position = extra[0]
+            self._discrepant_vcf_position.set_index("rsid", inplace=True)
+            logger.warning(
+                "{} SNP positions were found to be discrepant when saving VCF".format(
+                    len(self.discrepant_vcf_position)
+                )
+            )
+
+        return path
+
     def _read_raw_data(self, file, only_detect_source, rsids):
         return Reader.read_file(file, only_detect_source, self._resources, rsids)
 
@@ -485,13 +716,13 @@ class SNPs:
 
         Notes
         -----
-        rs3094315 : plus strand in 36, 37, and 38
-        rs11928389 : plus strand in 36, minus strand in 37 and 38
-        rs2500347 : plus strand in 36 and 37, minus strand in 38
-        rs964481 : plus strand in 36, 37, and 38
-        rs2341354 : plus strand in 36, 37, and 38
-        rs3850290 : plus strand in 36, 37, and 38
-        rs1329546 : plus strand in 36, 37, and 38
+        * rs3094315 : plus strand in 36, 37, and 38
+        * rs11928389 : plus strand in 36, minus strand in 37 and 38
+        * rs2500347 : plus strand in 36 and 37, minus strand in 38
+        * rs964481 : plus strand in 36, 37, and 38
+        * rs2341354 : plus strand in 36, 37, and 38
+        * rs3850290 : plus strand in 36, 37, and 38
+        * rs1329546 : plus strand in 36, 37, and 38
 
         Returns
         -------
@@ -514,7 +745,7 @@ class SNPs:
 
         def lookup_build_with_snp_pos(pos, s):
             try:
-                return s.loc[s == pos].index[0]
+                return int(s.loc[s == pos].index[0])
             except:
                 return 0
 
@@ -573,24 +804,7 @@ class SNPs:
 
         return build
 
-    def get_assembly(self):
-        """ Get the assembly of a build.
-
-        Returns
-        -------
-        str
-        """
-
-        if self._build == 37:
-            return "GRCh37"
-        elif self._build == 36:
-            return "NCBI36"
-        elif self._build == 38:
-            return "GRCh38"
-        else:
-            return ""
-
-    def get_snp_count(self, chrom=""):
+    def get_count(self, chrom=""):
         """ Count of SNPs.
 
         Parameters
@@ -607,57 +821,6 @@ class SNPs:
         else:
             return len(self._snps)
 
-    def get_chromosomes(self):
-        """ Get the chromosomes of SNPs.
-
-        Returns
-        -------
-        list
-            list of str chromosomes (e.g., ['1', '2', '3', 'MT'], empty list if no chromosomes
-        """
-
-        if not self._snps.empty:
-            return list(pd.unique(self._snps["chrom"]))
-        else:
-            return []
-
-    def get_chromosomes_summary(self):
-        """ Summary of the chromosomes of SNPs.
-
-        Returns
-        -------
-        str
-            human-readable listing of chromosomes (e.g., '1-3, MT'), empty str if no chromosomes
-        """
-
-        if not self._snps.empty:
-            chroms = list(pd.unique(self._snps["chrom"]))
-
-            int_chroms = [int(chrom) for chrom in chroms if chrom.isdigit()]
-            str_chroms = [chrom for chrom in chroms if not chrom.isdigit()]
-
-            # https://codereview.stackexchange.com/a/5202
-            def as_range(iterable):
-                l = list(iterable)
-                if len(l) > 1:
-                    return "{0}-{1}".format(l[0], l[-1])
-                else:
-                    return "{0}".format(l[0])
-
-            # create str representations
-            int_chroms = ", ".join(
-                as_range(g)
-                for _, g in groupby(int_chroms, key=lambda n, c=count(): n - next(c))
-            )
-            str_chroms = ", ".join(str_chroms)
-
-            if int_chroms != "" and str_chroms != "":
-                int_chroms += ", "
-
-            return int_chroms + str_chroms
-        else:
-            return ""
-
     def determine_sex(
         self,
         heterozygous_x_snps_threshold=0.03,
@@ -688,10 +851,10 @@ class SNPs:
         return ""
 
     def _determine_sex_X(self, threshold):
-        x_snps = self.get_snp_count("X")
+        x_snps = self.get_count("X")
 
         if x_snps > 0:
-            if len(self.heterozygous_snps("X")) / x_snps > threshold:
+            if len(self.heterozygous("X")) / x_snps > threshold:
                 return "Female"
             else:
                 return "Male"
@@ -699,10 +862,10 @@ class SNPs:
             return ""
 
     def _determine_sex_Y(self, threshold):
-        y_snps = self.get_snp_count("Y")
+        y_snps = self.get_count("Y")
 
         if y_snps > 0:
-            if len(self.not_null_snps("Y")) / y_snps > threshold:
+            if len(self.notnull("Y")) / y_snps > threshold:
                 return "Male"
             else:
                 return "Female"
@@ -744,19 +907,23 @@ class SNPs:
         # Keep first duplicate rsid.
         duplicate_rsids = self._snps.index.duplicated(keep="first")
         # save duplicate SNPs
-        self._duplicate_snps = self._duplicate_snps.append(
-            self._snps.loc[duplicate_rsids]
-        )
+        self._duplicate = self._duplicate.append(self._snps.loc[duplicate_rsids])
         # deduplicate
         self._snps = self._snps.loc[~duplicate_rsids]
 
-    def _deduplicate_chrom(self, chrom):
+    def _deduplicate_alleles(self, rsids):
+        # remove duplicate allele
+        self._snps.loc[rsids, "genotype"] = self._snps.loc[rsids, "genotype"].apply(
+            lambda x: x[0]
+        )
+
+    def _deduplicate_sex_chrom(self, chrom):
         """ Deduplicate a chromosome in the non-PAR region. """
 
         discrepant_XY_snps = self._get_non_par_snps(chrom)
 
         # save discrepant XY SNPs
-        self._discrepant_XY_snps = self._discrepant_XY_snps.append(
+        self._discrepant_XY = self._discrepant_XY.append(
             self._snps.loc[discrepant_XY_snps]
         )
 
@@ -766,15 +933,26 @@ class SNPs:
         # get remaining non-PAR SNPs with two alleles
         non_par_snps = self._get_non_par_snps(chrom, heterozygous=False)
 
-        # remove duplicate allele
-        self._snps.loc[non_par_snps, "genotype"] = self._snps.loc[
-            non_par_snps, "genotype"
-        ].apply(lambda x: x[0])
+        self._deduplicate_alleles(non_par_snps)
 
     def _deduplicate_XY_chrom(self):
         """ Fix chromosome issue where some data providers duplicate male X and Y chromosomes"""
-        self._deduplicate_chrom("X")
-        self._deduplicate_chrom("Y")
+        self._deduplicate_sex_chrom("X")
+        self._deduplicate_sex_chrom("Y")
+
+    def _deduplicate_MT_chrom(self):
+        """ Deduplicate MT chromosome. """
+        heterozygous_MT_snps = self._snps.loc[self.heterozygous("MT").index].index
+
+        # save heterozygous MT SNPs
+        self._heterozygous_MT = self._heterozygous_MT.append(
+            self._snps.loc[heterozygous_MT_snps]
+        )
+
+        # drop heterozygous MT SNPs since it's ambiguous for which allele to deduplicate
+        self._snps.drop(heterozygous_MT_snps, inplace=True)
+
+        self._deduplicate_alleles(self.homozygous("MT").index)
 
     @staticmethod
     def get_par_regions(build):
@@ -830,7 +1008,7 @@ class SNPs:
         else:
             return pd.DataFrame()
 
-    def sort_snps(self):
+    def sort(self):
         """ Sort SNPs based on ordered chromosome list and position. """
 
         sorted_list = sorted(self._snps["chrom"].unique(), key=self._natural_sort_key)
@@ -858,7 +1036,7 @@ class SNPs:
 
         self._snps = snps
 
-    def remap_snps(self, target_assembly, complement_bases=True):
+    def remap(self, target_assembly, complement_bases=True):
         """ Remap SNP coordinates from one assembly to another.
 
         This method uses the assembly map endpoint of the Ensembl REST API service (via
@@ -969,9 +1147,10 @@ class SNPs:
         # update SNP positions and genotypes
         snps.loc[remapped_snps.index, "pos"] = remapped_snps["pos"]
         snps.loc[remapped_snps.index, "genotype"] = remapped_snps["genotype"]
+        snps.pos = snps.pos.astype(np.uint32)
 
         self._snps = snps
-        self.sort_snps()
+        self.sort()
         self._build = int(target_assembly[-2:])
 
         return chromosomes_remapped, chromosomes_not_remapped
@@ -1081,3 +1260,348 @@ class SNPs:
             int(text) if text.isdigit() else text.lower()
             for text in re.split(natural_sort_re, s)
         ]
+
+    def merge(
+        self,
+        snps_objects=(),
+        discrepant_positions_threshold=100,
+        discrepant_genotypes_threshold=500,
+        remap=True,
+    ):
+        """ Merge other ``SNPs`` objects into this ``SNPs`` object.
+
+        Parameters
+        ----------
+        snps_objects : list or tuple of ``SNPs``
+            other ``SNPs`` objects to merge into this ``SNPs`` object
+        discrepant_positions_threshold : int
+            threshold for discrepant SNP positions between existing data and data to be loaded;
+            a large value could indicate mismatched genome assemblies
+        discrepant_genotypes_threshold : int
+            threshold for discrepant genotype data between existing data and data to be loaded;
+            a large value could indicated mismatched individuals
+        remap : bool
+            if necessary, remap other ``SNPs`` objects to have the same build as this ``SNPs`` object
+            before merging
+
+        Returns
+        -------
+        list of dict
+            for each ``SNPs`` object to merge, a dict with the following items:
+
+            merged (bool)
+                whether ``SNPs`` object was merged
+            common_rsids (pandas.Index)
+                SNPs in common
+            discrepant_position_rsids (pandas.Index)
+                SNPs with discrepant positions
+            discrepant_genotype_rsids (pandas.Index)
+                SNPs with discrepant genotypes
+
+        References
+        ----------
+        1. Fluent Python by Luciano Ramalho (O'Reilly). Copyright 2015 Luciano Ramalho,
+           978-1-491-94600-8.
+        """
+
+        def init(s):
+            # initialize this SNPs object with properties of the SNPs object being merged
+            self._snps = s.snps
+            self._duplicate = s.duplicate
+            self._discrepant_XY = s.discrepant_XY
+            self._heterozygous_MT = s.heterozygous_MT
+            self._discrepant_vcf_position = s.discrepant_vcf_position
+            self._discrepant_merge_positions = s.discrepant_merge_positions
+            self._discrepant_merge_genotypes = s.discrepant_merge_genotypes
+            self._source = s._source
+            self._phased = s.phased
+            self._build = s.build
+            self._build_detected = s.build_detected
+
+        def ensure_same_build(s):
+            # ensure builds match when merging
+            if not s.build_detected:
+                logger.warning(
+                    "Build not detected for {}, assuming Build {}".format(
+                        s.__repr__(), s.build
+                    )
+                )
+
+            if self.build != s.build:
+                logger.info(
+                    "{} has Build {}; remapping to Build {}".format(
+                        s.__repr__(), s.build, self.build
+                    )
+                )
+                s.remap(self.build)
+
+        def merge_properties(s):
+            if not s.build_detected:
+                # can no longer assume build has been detected for all SNPs after merge
+                self._build_detected = False
+
+            if not s.phased:
+                # can no longer assume all SNPs are phased after merge
+                self._phased = False
+
+            self._source.extend(s._source)
+
+        def merge_dfs(s):
+            # append dataframes created when a ``SNPs`` object is instantiated
+            self._duplicate = self.duplicate.append(s.duplicate)
+            self._discrepant_XY = self.discrepant_XY.append(s.discrepant_XY)
+            self._heterozygous_MT = self.heterozygous_MT.append(s.heterozygous_MT)
+            self._discrepant_vcf_position = self.discrepant_vcf_position.append(
+                s.discrepant_vcf_position
+            )
+
+        def merge_snps(s, positions_threshold, genotypes_threshold):
+            # merge SNPs, identifying those with discrepant positions and genotypes; update NAs
+
+            # identify common SNPs (i.e., any rsids being added that already exist in self.snps)
+            df = self.snps.join(s.snps, how="inner", rsuffix="_added")
+
+            common_rsids = df.index
+
+            discrepant_positions = df.loc[
+                (df.chrom != df.chrom_added) | (df.pos != df.pos_added)
+            ]
+
+            if len(discrepant_positions) >= positions_threshold:
+                logger.warning(
+                    "Too many SNPs differ in position; ensure SNPs have same build"
+                )
+                return (False,)
+
+            # remove null genotypes
+            df = df.loc[~df.genotype.isnull() & ~df.genotype_added.isnull()]
+
+            # discrepant genotypes are where alleles are not equivalent (i.e., alleles are not the
+            # same and not swapped)
+            discrepant_genotypes = df.loc[
+                (df.genotype.str.len() != df.genotype_added.str.len())
+                | (
+                    (df.genotype.str.len() == 1)
+                    & (df.genotype_added.str.len() == 1)
+                    & (df.genotype != df.genotype_added)
+                )
+                | (
+                    (df.genotype.str.len() == 2)
+                    & (df.genotype_added.str.len() == 2)
+                    & ~(
+                        (df.genotype.str[0] == df.genotype_added.str[0])
+                        & (df.genotype.str[1] == df.genotype_added.str[1])
+                    )
+                    & ~(
+                        (df.genotype.str[0] == df.genotype_added.str[1])
+                        & (df.genotype.str[1] == df.genotype_added.str[0])
+                    )
+                )
+            ]
+
+            if len(discrepant_genotypes) >= genotypes_threshold:
+                logger.warning(
+                    "Too many SNPs differ in their genotype; ensure file is for same "
+                    "individual"
+                )
+                return (False,)
+
+            # add new SNPs
+            self._snps = self.snps.combine_first(s.snps)
+            # combine_first converts position to float64, so convert it back to uint32
+            self._snps["pos"] = self.snps["pos"].astype(np.uint32)
+
+            if 0 < len(discrepant_positions) < positions_threshold:
+                logger.warning(
+                    "{} SNP positions were discrepant; keeping original positions".format(
+                        str(len(discrepant_positions))
+                    )
+                )
+
+            if 0 < len(discrepant_genotypes) < genotypes_threshold:
+                logger.warning(
+                    "{} SNP genotypes were discrepant; marking those as null".format(
+                        str(len(discrepant_genotypes))
+                    )
+                )
+
+            # set discrepant genotypes to null
+            self._snps.loc[discrepant_genotypes.index, "genotype"] = np.nan
+
+            # append discrepant positions dataframe
+            self._discrepant_merge_positions = self._discrepant_merge_positions.append(
+                discrepant_positions, sort=True
+            )
+            # append discrepant genotypes dataframe
+            self._discrepant_merge_genotypes = self._discrepant_merge_genotypes.append(
+                discrepant_genotypes, sort=True
+            )
+
+            return (
+                True,
+                {
+                    "common_rsids": common_rsids,
+                    "discrepant_position_rsids": discrepant_positions.index,
+                    "discrepant_genotype_rsids": discrepant_genotypes.index,
+                },
+            )
+
+        results = []
+        for snps_object in snps_objects:
+            d = {
+                "merged": False,
+                "common_rsids": pd.Index([], name="rsid"),
+                "discrepant_position_rsids": pd.Index([], name="rsid"),
+                "discrepant_genotype_rsids": pd.Index([], name="rsid"),
+            }
+
+            if not snps_object.valid:
+                logger.warning("No SNPs to merge...")
+                results.append(d)
+                continue
+
+            if not self.valid:
+                logger.info("Loading {}".format(snps_object.__repr__()))
+
+                init(snps_object)
+                d.update({"merged": True})
+            else:
+                logger.info("Merging {}".format(snps_object.__repr__()))
+
+                if remap:
+                    ensure_same_build(snps_object)
+
+                if self.build != snps_object.build:
+                    logger.warning(
+                        "{} has Build {}; this SNPs object has Build {}".format(
+                            snps_object.__repr__(), snps_object.build, self.build
+                        )
+                    )
+
+                merged, *extra = merge_snps(
+                    snps_object,
+                    discrepant_positions_threshold,
+                    discrepant_genotypes_threshold,
+                )
+
+                if merged:
+                    merge_properties(snps_object)
+                    merge_dfs(snps_object)
+                    self.sort()
+
+                    d.update({"merged": True})
+                    d.update(extra[0])
+
+            results.append(d)
+
+        return results
+
+    def sort_snps(self):
+        warnings.warn("This method has been renamed to `sort`.", DeprecationWarning)
+        self.sort()
+
+    def remap_snps(self, target_assembly, complement_bases=True):
+        warnings.warn("This method has been renamed to `remap`.", DeprecationWarning)
+        return self.remap(target_assembly, complement_bases)
+
+    def save_snps(self, filename="", vcf=False, atomic=True, **kwargs):
+        warnings.warn("This method has been renamed to `save`.", DeprecationWarning)
+        return self.save(filename, vcf, atomic, **kwargs)
+
+    @property
+    def snp_count(self):
+        warnings.warn("This property has been renamed to `count`.", DeprecationWarning)
+        return self.count
+
+    def get_snp_count(self, chrom=""):
+        warnings.warn(
+            "This method has been renamed to `get_count`.", DeprecationWarning
+        )
+        return self.get_count(chrom)
+
+    def not_null_snps(self, chrom=""):
+        warnings.warn("This method has been renamed to `notnull`.", DeprecationWarning)
+        return self.notnull(chrom)
+
+    def get_summary(self):
+        warnings.warn(
+            "This method has been renamed to `summary` and is now a property.",
+            DeprecationWarning,
+        )
+        return self.summary
+
+    def get_assembly(self):
+        warnings.warn("See the `assembly` property.", DeprecationWarning)
+        return self.assembly
+
+    def get_chromosomes(self):
+        warnings.warn("See the `chromosomes` property.", DeprecationWarning)
+        return self.chromosomes
+
+    def get_chromosomes_summary(self):
+        warnings.warn("See the `chromosomes_summary` property.", DeprecationWarning)
+        return self.chromosomes_summary
+
+    @property
+    def duplicate_snps(self):
+        warnings.warn(
+            "This property has been renamed to `duplicate`.", DeprecationWarning
+        )
+        return self.duplicate
+
+    @property
+    def discrepant_XY_snps(self):
+        warnings.warn(
+            "This property has been renamed to `discrepant_XY`.", DeprecationWarning
+        )
+        return self.discrepant_XY
+
+    @property
+    def heterozygous_MT_snps(self):
+        warnings.warn(
+            "This property has been renamed to `heterozygous_MT`.", DeprecationWarning
+        )
+        return self.heterozygous_MT
+
+    def heterozygous_snps(self, chrom=""):
+        warnings.warn(
+            "This method has been renamed to `heterozygous`.", DeprecationWarning
+        )
+        return self.heterozygous(chrom)
+
+    def homozygous_snps(self, chrom=""):
+        warnings.warn(
+            "This method has been renamed to `homozygous`.", DeprecationWarning
+        )
+        return self.homozygous(chrom)
+
+    @property
+    def discrepant_positions(self):
+        warnings.warn(
+            "This property has been renamed to `discrepant_merge_positions`.",
+            DeprecationWarning,
+        )
+        return self.discrepant_merge_positions
+
+    @property
+    def discrepant_genotypes(self):
+        warnings.warn(
+            "This property has been renamed to `discrepant_merge_genotypes`.",
+            DeprecationWarning,
+        )
+        return self.discrepant_merge_genotypes
+
+    @property
+    def discrepant_snps(self):
+        warnings.warn(
+            "This property has been renamed to `discrepant_merge_positions_genotypes`.",
+            DeprecationWarning,
+        )
+        return self.discrepant_merge_positions_genotypes
+
+    def is_valid(self):
+        warnings.warn(
+            "This method has been renamed to `valid` and is now a property.",
+            DeprecationWarning,
+        )
+        return self.valid
